;;;; objc-base.scm


(declare
  (uses srfi-1)
  (export objc:nslog
	  objc:instance? objc:selector? objc:class? objc:instance->pointer objc:pointer->instance 
	  objc:pointer->class
	  objc:resolve-with-cache objc:nsstring
	  objc:string->class objc:string->selector objc:selector->signature) )

(use lazy-ffi)

#>
#import <Foundation/Foundation.h>
#import <objc/objc-runtime.h>

static NSAutoreleasePool *default_pool = NULL;
<#

(define objc:nslog
  (let ((nslog (foreign-lambda* void ((c-string s)) 
		 "NSString *str = [[NSString alloc] initWithUTF8String: s];"
		 "NSLog(str);"
		 "[str release];") ) )
    (lambda (fstr . args)
      (nslog (apply sprintf fstr args)) ) ) )

#>?
___declare(default_renaming, "")
<#

#>:
default_pool = [[NSAutoreleasePool alloc] init];
<#

#>!
static const char *objc_description(void *s) 
{
  return [[(NSObject *)s description] UTF8String];
}

static void *string_to_class(char *name)
{
  NSString *str = [[NSString alloc] initWithUTF8String: name];
  void *clazz = (void *)NSClassFromString(str);
  [str release];
  return clazz;
}

static void *string_to_selector(char *name)
{
  NSString *str = [[NSString alloc] initWithUTF8String: name];
  SEL s = (void *)NSSelectorFromString(str);
  [str release];
  return (void *)s;
}

static const char *selector_chars(void *sel)
{
  SEL s = (SEL)sel;
  return (char *)s;
}

static void *selector_to_signature(void *x, void *sel, ___bool clazz)
{
  NSObject *o = (NSObject *)x;

  if(clazz) return (void *)[o methodSignatureForSelector: (SEL)sel];
  else return (void *)[[o class] instanceMethodSignatureForSelector: (SEL)sel];
}

static const char *method_return_type(void *sig)
{
  NSMethodSignature *s = (NSMethodSignature *)sig;
  return [s methodReturnType];
}

static int method_argument_count(void *sig)
{
  NSMethodSignature *s = (NSMethodSignature *)sig;
  return [s numberOfArguments];
}

static const char *method_argument_type(void *sig, int i)
{
  NSMethodSignature *s = (NSMethodSignature *)sig;
  return [s getArgumentTypeAtIndex: i];
}

static void *make_nsstring(char *s)
{
  return [[NSString alloc] initWithUTF8String: s];
}

static ___bool is_nsstring(void *o)
{
  NSObject *ob = (NSObject *)o;
  return [ob isKindOfClass: [NSString class]];
}

static void objc_release(void *o)
{
  NSObject *obj = (NSObject *)o;
  [obj release];
}
<#

(define-record objc:instance ptr)
(define-record objc:class ptr)
(define-record objc:selector ptr)

(define (objc:nsstring s)
  (make-objc:instance (make-nsstring s) ))

(define-record-printer (objc:instance ptr p)
  (let ((ptr (objc:instance-ptr ptr)))
    (if (is-nsstring ptr)
	(fprintf p "@~s" (objc-description ptr))
	(fprintf p "#<objc-instance ~a>" (objc-description ptr))) ) ) 

(define-record-printer (objc:class ptr p)
  (fprintf p "#<objc-class ~a>" (objc-description (objc:class-ptr ptr))) )

(define-record-printer (objc:selector ptr p)
  (fprintf p "#<objc-selector ~a>" (selector-chars (objc:selector-ptr ptr))) )

(define (objc:string->class str)
  (make-objc:class (string-to-class str)) )

(define (objc:string->selector str)
  (make-objc:selector (string-to-selector str)))

(define (objc:instance->pointer x)
  (and x
       (if (objc:class? x)
	   (objc:class-ptr x)
	   (objc:instance-ptr x) ) ) )

(define (objc:pointer->instance p)
  (make-objc:instance p) )

(define (objc:pointer->class p)
  (make-objc:class p) )

(define (objc:selector->signature o sel)
  (let* ((c? (objc:class? o))
	 (ptr (if c? (objc:class-ptr o) (objc:instance-ptr o))) )
    (make-objc:instance (selector-to-signature ptr (objc:selector-ptr sel) c?) ) ) )

(define (objc:method-signature o sel)
  (let ((sig (selector-to-signature (objc:instance->pointer o) (objc:selector-ptr sel) (objc:class? o))))
    (and sig
	 (let* ((n (method-argument-count sig))
		(slist
		 (cons (method-return-type sig)
		       (let loop ((j 2))
			 (if (fx>= j n)
			     '()
			     (cons (method-argument-type sig j)
				   (loop (fx+ j 1)) ) ) ) ) ) )
	   (objc-release sig)
	   slist) ) ) )

(define objc_msgSend (lazy-ffi:function-ptr (foreign-value "(void *)objc_msgSend" c-pointer)))

(define (unwrap-obj x)
  (and x (objc:instance-ptr x)) )

(define (wrap-obj x)
  (and x (make-objc:instance x)))

(define (unwrap-cls x)
  (and x (objc:class-ptr x)) )

(define (wrap-cls x)
  (and x (make-objc:class x)))

(define-foreign-type objcinst c-pointer unwrap-obj wrap-obj)
(define-foreign-type objccls c-pointer unwrap-cls wrap-cls)

(define simple-calls 
  (alist->hash-table
   `((retain  . ,(lambda (ro) ((foreign-lambda* void ((objcinst x)) "[((NSObject *)x) retain];") ro)))
     (alloc  . ,(lambda (ro) ((foreign-lambda* objcinst ((objccls x)) "return([((Class)x) alloc]);") ro)))
     (init  . ,(lambda (ro) ((foreign-lambda* objcinst ((objcinst x)) "return([((NSObject *)x) init]);") ro)))
     (release  . ,(lambda (ro) ((foreign-lambda* void ((objcinst x)) "[((NSObject *)x) release];") ro)))
     (autorelease  . ,(lambda (ro) ((foreign-lambda* void ((objcinst x)) "[((NSObject *)x) autorelease];") ro)))
     (length  . ,(lambda (ro) ((foreign-lambda* int ((objcinst x)) "return([((NSString *)x) length]);") ro)))
     (class  . ,(lambda (ro) ((foreign-lambda* objccls ((objcinst x)) "return([((NSObject *)x) class]);") ro)))
     (description  . ,(lambda (ro) ((foreign-lambda* objcinst ((objcinst x)) "return([((NSObject *)x) description]);") ro)))
     (cString  . ,(lambda (ro) ((foreign-lambda* c-string ((objcinst x)) "return([((NSString *)x) cString]);") ro)))
     (UTF8String  . ,(lambda (ro) ((foreign-lambda* c-string ((objcinst x)) "return([((NSString *)x) UTF8String]);") ro)))
     (self  . ,(lambda (ro) ((foreign-lambda* objcinst ((objcinst x)) "return([((NSObject *)x) self]);") ro)))
     (dealloc  . ,(lambda (ro) ((foreign-lambda* void ((objcinst x)) "[((NSObject *)x) dealloc];") ro))) )
   eq?))

(define (objc:resolve-with-cache cache keysym receiver keys safe?)
  (define (resolve c)
    (let* ((sel (objc:string->selector keys))
	   (sig (objc:method-signature receiver sel))
	   (call
	    (lambda (args)
	      (objc:send-message-0 receiver sel sig args safe?) ) ) )
      (##sys#setslot c 0 (objc:instance->pointer receiver))
      (##sys#setslot c 1 keysym)
      (##sys#setslot c 2 call)
      call) )
  (and receiver
       (let ((so (hash-table-ref/default simple-calls keysym #f))) ; should use symbol/internal hash-tables
	 (cond (so
		#+debug (printf "[simple message: ~s]~%" keysym)
		(lambda _ (so receiver)) )
	       ((##sys#symbol-has-toplevel-binding? cache)
		(let ((c (##sys#slot cache 0)))
		  (if (and (equal? (##sys#slot c 0) (objc:instance-ptr receiver))
			   (eq? keysym (##sys#slot c 1)) )
		      (begin
			#+debug (printf "[cache hit for: ~s]~%" keysym)
			(##sys#slot c 2) )
		      (resolve c) ) ) )
	       (else
		(let ((c (make-vector 3)))
		  (##sys#setslot cache 0 c)
		  (resolve c)))))))

(define objc:send-message-0
  (let ()
    (define (rtype ot)
      (let loop ((i 0))
	(case (string-ref ot i)
	  ((#\c) char:)
	  ((#\v) void:)
	  ((#\s #\i #\l #\C #\I #\S #\L) int:)
	  ((#\f) float:)
	  ((#\d) double:)
	  ((#\*) string:)
	  ((#\@ #\: #\^ #\#) pointer:)
	  ((#\r #\n #\N #\R #\V #\o #\O) (loop (fx+ i 1)))
	  (else (error 'objc:send-message-0 "can not convert result type" ot)) ) ) )
    (lambda (ro sel sig args safe?)
      (let ((rt (rtype (car sig)))
	    (rop (objc:instance->pointer ro))
	    (sptr (objc:selector-ptr sel)) )
	(let ((result
	       (apply 
		objc_msgSend
		safe: safe?
		return: rt
		pointer: rop
		pointer: sptr
		(let loop ((args args) (types (cdr sig)))
		  (cond ((null? types)
			 (if (null? args)
			     '()
			     (error 'objc:send-message-0 "too many arguments in method call" ro sel args) ) )
			((null? args)	
			 (error 'objc:send-message-0 "too few arguments in method call" ro sel args) )
			(else
			 (cons* (rtype (car types)) 
				(let ((arg (car args)))
				  (cond ((objc:instance? arg) (objc:instance-ptr arg))
					((objc:class? arg) (objc:class-ptr arg))
					((objc:selector? arg) (objc:selector-ptr arg))
					(else arg) ) )
				(loop (cdr args) (cdr types)))))))))
	  (case (string-ref (car sig) 0)
	    ((#\@) (make-objc:instance result))
	    ((#\#) (make-objc:class result))
	    ((#\:) (make-objc:selector result))
	    (else result) ) ) ) ) ) )
