;;;; objc.scm


(require 'lazy-ffi)

(define-macro (define-objc-classes . names)
  `(begin
     ,@(map (lambda (name)
	      (cond ((symbol? name)
		     `(define ,name (objc:string->class ,(->string name)) ))
		    ((and (list? name) (= (length name) 2) (symbol? (car name)))
		     `(define ,(car name) (objc:string->class ,(->string (cadr name)))) )
		    (else (syntax-error 'define-objc-classes "invalid class name" name))))
	    names) ) )

(set-read-syntax! 
 #\@
 (let ([terminating-characters '(#\, #\; #\) #\] #\{ #\} #\')])
   (lambda (p)
     (let ((c (peek-char p)))
       (if (or (char-whitespace? c) (memq c terminating-characters))
	   '@
	   (let ((x (read p)))
	     (cond ((keyword? x) (string->keyword (string-append "@" (keyword->string x))))
		   ((symbol? x) (string->symbol (string-append "@" (symbol->string x))))
		   ((string? x) `(force (delay (objc:nsstring ,x))))
		   ((pair? x) 
		    (if (eq? #:safe (car x))
			`(objc:send/safe ,@(cdr x))
			`(objc:send ,@x)))
		   (else (error "invalid read syntax for `@'" c)) ) ) ) ) ) ) )

(eval-when (compile eval)
  (set! objc:decompose-message
    (lambda (sym1 syms loc)
      (let ((args (cons sym1 syms)))
	(if (and (null? syms) (not (keyword? sym1)) (symbol? sym1))
	    (values (symbol->string sym1) '())
	    (let loop ((msg args) (keys '()) (vals '()))
	      (if (null? msg) 
		  (values (apply string-append (reverse keys)) (reverse vals))
		  (let ((sel (car msg))
			(r (cdr msg)) )
		    (cond ((keyword? sel)
			   (if (null? r)
			       (syntax-error loc "invalid method-invocation syntax - missing arguments" args)
			       (loop (cdr r) (cons ":" (cons (keyword->string sym1) keys)) (cons (car r) vals))))
			  (else (syntax-error loc "invalid selector syntax - bad keyword" sel) ) ) ) ) ) ) ) ) ) )

(define-macro (objc:send receiver sym1 . msg)
  (let-values (((keys vals) (objc:decompose-message sym1 msg 'objc:send)))
    `((objc:resolve-with-cache ',(lazy-ffi:generate-uid sym1) ',(string->symbol keys) ,receiver ,keys #f) (list ,@vals))))

(define-macro (objc:send/safe receiver sym1 . msg)
  (let-values (((keys vals) (objc:decompose-message sym1 msg 'objc:send)))
    `((objc:resolve-with-cache ',(lazy-ffi:generate-uid sym1) ',(string->symbol keys) ,receiver ,keys #t) (list ,@vals))))
